/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AMRNavierStokes.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/


#include "AMRNavierStokesHeaders.H"




void 
AMRNavierStokes::fillLambda(LevelData<FArrayBox>& a_lambda, Real a_time) const
{
  assert (!a_lambda.isDefined());
  
  IntVect ghostVect(D_DECL(ADVECT_GROW, ADVECT_GROW, ADVECT_GROW));
  const DisjointBoxLayout& levelGrids = m_vel_new_ptr->getBoxes();
  int ncomp = 1;

  a_lambda.define(levelGrids, ncomp, ghostVect);

  if (s_set_bogus_values) {
    DataIterator dit = a_lambda.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      a_lambda[dit()].setVal(s_bogus_value);
    }
  }

  Interval lambdaComps(0, ncomp-1);

  Real old_time = m_time - m_dt;
  // first copy lambda to a_lambda
  if (abs(a_time - old_time) < TIME_EPS) {
    // copy from old values
    m_lambda_old_ptr->copyTo(lambdaComps, a_lambda, lambdaComps);
  } else if (abs(a_time - m_time) < TIME_EPS) {
    // copy from new values
    m_lambda_new_ptr->copyTo(lambdaComps, a_lambda, lambdaComps);
  } else {
    // interp in time
    timeInterp(a_lambda, a_time, oldLambda(), old_time,
	       newLambda(), m_time, lambdaComps);
  }

  if (m_level > 0) {
    // interpolate boundary values from coarse grids
    AMRNavierStokes* crseLevelPtr = crseNSPtr();
    assert (crseLevelPtr != NULL);
    LevelData<FArrayBox>& crse_old_lambda = crseLevelPtr->oldLambda();
    LevelData<FArrayBox>& crse_new_lambda = crseLevelPtr->newLambda();
    const DisjointBoxLayout& crseGrids = crse_new_lambda.getBoxes();
    const ProblemDomain& crseDomain = crseLevelPtr->problemDomain();
    int nRefCrse = crseLevelPtr->refRatio();

    Real crse_dt = crseLevelPtr->dt();
    // crse time should be the new time    
    Real crse_old_time = crseLevelPtr->time() - crse_dt;
    Real time_interp_coeff;

    // check for "essentially 0 or 1"
    if (abs(a_time - crse_old_time) < TIME_EPS) {
      time_interp_coeff = 0.0;
    } else if (abs(a_time - crseLevelPtr->time()) < TIME_EPS) {
      time_interp_coeff = 1.0;
    } else {
      time_interp_coeff = (a_time - crse_old_time)/crse_dt;
    }




    PiecewiseLinearFillPatch filpatcher(levelGrids, crseGrids,
					ncomp, crseDomain, nRefCrse,
					ADVECT_GROW);

    
    filpatcher.fillInterp(a_lambda, crse_old_lambda, crse_new_lambda, 
			  time_interp_coeff, 0, 0, ncomp);

  }

  // now do physical boundary conditions
  DomainGhostBC* lambdaBCPtr = m_physBCPtr->lambdaBC();

  // loop over boxes
  const ProblemDomain& physDomain = problemDomain();
  DataIterator dit = a_lambda.dataIterator();
  for (dit.reset(); dit.ok(); ++dit) {
    lambdaBCPtr->applyInhomogeneousBCs(a_lambda[dit()], physDomain, m_dx);
  }

  delete lambdaBCPtr;

}
  

void
AMRNavierStokes::fillVelocity(LevelData<FArrayBox>& a_vel, Real a_time)
{

#if 0
  // this gets called so often it was getting annoying
  if (s_verbosity >= 3) {
    
    pout() << "AMRNavierStokes::fillVelocity(" << a_time << "): " 
	   << m_level << endl;
  }  
#endif  

  Interval velComps(0,SpaceDim-1);

  if (s_set_bogus_values) {
    DataIterator dit = a_vel.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      a_vel[dit()].setVal(s_bogus_value);
    }
  }

  Real old_time = m_time - m_dt;

  // first copy on this level
  if (abs(a_time - old_time) < TIME_EPS) {
    m_vel_old_ptr->copyTo(velComps, a_vel, velComps);
  } else if (abs(a_time - m_time) < TIME_EPS) {
    m_vel_new_ptr->copyTo(velComps, a_vel, velComps);
  } else {
    // do linear interpolation in time
    timeInterp(a_vel, a_time, oldVel(), old_time,
	       newVel(), m_time, velComps);
  }


  
  // if necessary, do interpolation from coarser levels
  if (m_level > 0) {
    const DisjointBoxLayout& levelGrids = a_vel.getBoxes();
    const DisjointBoxLayout& thisLevelsGrids = newVel().getBoxes();

    const IntVect& velGrowVect = a_vel.ghostVect();
    int velGrow = velGrowVect[0];
    
    // if grids for a_vel are the same as those for this level, and 
    // there are no ghost cells, we don't need to do this at all
    if (!( (velGrow == 0) && (levelGrids == thisLevelsGrids))) {
      
      // fill in coarse-fine BC data by conservative linear interp
      AMRNavierStokes& crseLevel = *crseNSPtr();
      
      LevelData<FArrayBox>& oldCrseVel = crseLevel.oldVel();
      LevelData<FArrayBox>& newCrseVel = crseLevel.newVel();
      const DisjointBoxLayout& crseGrids = oldCrseVel.getBoxes();
      const ProblemDomain& crseDomain = crseLevel.problemDomain();
      int nRefCrse = crseLevel.refRatio();
      
      Real crse_new_time = crseLevel.m_time;
      Real crse_dt = crseLevel.dt();
      Real crse_old_time = crse_new_time - crse_dt;
      Real crse_time_interp_coeff;
      
      // check for "essentially 0 or 1"
      if (abs(a_time - crse_old_time) < TIME_EPS) {
        crse_time_interp_coeff = 0.0;
      } else if (abs(a_time - crse_new_time) < TIME_EPS) {
        crse_time_interp_coeff = 1.0;
      } else {
        crse_time_interp_coeff = (a_time - crse_old_time)/crse_dt;
      }
      
      
      
      PiecewiseLinearFillPatch filpatcher(levelGrids, crseGrids, 
                                          SpaceDim, crseDomain,
                                          nRefCrse, velGrow);
      
      
      filpatcher.fillInterp(a_vel, oldCrseVel, newCrseVel,
                            crse_time_interp_coeff, 0,0,SpaceDim);
    }
  }
  
  // need to set physical boundary conditions here
  basicVelBC* velBCPtr;
  if (s_nu > 0.0) {
    velBCPtr = m_physBCPtr->viscousVelBC();
  } else {
    velBCPtr = m_physBCPtr->tracingVelBC();
  }

  velBCPtr->applyBCs(a_vel, m_problem_domain, m_dx);

  delete velBCPtr;

  // should probably put an exchange here
  a_vel.exchange(a_vel.interval());
}


void
AMRNavierStokes::fillVelocity(LevelData<FArrayBox>& a_vel, Real a_time,
			      int vel_comp, int dest_comp, int num_comp)
{

#if 0
  // this gets called way too often!
  if (s_verbosity >= 3) {
    pout() << "AMRNavierStokes::fillVelocity " << m_level << endl;
  }  
#endif

  Interval srcComps(vel_comp,vel_comp+num_comp-1);
  Interval destComps(dest_comp, dest_comp+num_comp-1);


  Real old_time = m_time - m_dt;
  // first copy on this level
  if (abs(a_time - old_time) < TIME_EPS) {
    m_vel_old_ptr->copyTo(srcComps, a_vel, destComps);
  } else if (abs(a_time - m_time) < TIME_EPS) {
    m_vel_new_ptr->copyTo(srcComps, a_vel, destComps);
  } else {
    // do linear interpolation in time
    timeInterp(a_vel, a_time, oldVel(), old_time,
	       newVel(), m_time, srcComps,destComps);
  }
  
  // if necessary, do interpolation from coarser levels
  if (m_level > 0) {
    const DisjointBoxLayout& levelGrids = a_vel.getBoxes();
    const DisjointBoxLayout& thisLevelsGrids = newVel().getBoxes();

    const IntVect& velGrowVect = a_vel.ghostVect();
    int velGrow = velGrowVect[0];
    
    // if grids for a_vel are the same as those for this level, and 
    // there are no ghost cells, we don't need to do this at all
    if (!( (velGrow == 0) && (levelGrids == thisLevelsGrids))) {

      const DisjointBoxLayout& levelGrids = a_vel.getBoxes();
      
      // fill in coarse-fine BC data by conservative linear interp
      AMRNavierStokes& crseLevel = *crseNSPtr();
      
      LevelData<FArrayBox>& oldCrseVel = crseLevel.oldVel();
      LevelData<FArrayBox>& newCrseVel = crseLevel.newVel();
      const DisjointBoxLayout& crseGrids = oldCrseVel.getBoxes();
      const ProblemDomain& crseDomain = crseLevel.problemDomain();
      int nRefCrse = crseLevel.refRatio();
      
      Real crse_new_time = crseLevel.m_time;
      Real crse_dt = crseLevel.dt();
      Real crse_old_time = crse_new_time - crse_dt;
      Real crse_time_interp_coeff;
      
      // check for "essentially 0 or 1"
      if (abs(a_time - crse_old_time) < TIME_EPS) {
        crse_time_interp_coeff = 0.0;
      } else if (abs(a_time - crse_new_time) < TIME_EPS) {
        crse_time_interp_coeff = 1.0;
      } else {
        crse_time_interp_coeff = (a_time - crse_old_time)/crse_dt;
      }
      
      
      
      const IntVect& velGrowVect = a_vel.ghostVect();
      int velGrow = velGrowVect[0];
      
      PiecewiseLinearFillPatch filpatcher(levelGrids, crseGrids, 
                                          num_comp, crseDomain,
                                          nRefCrse, velGrow);
      
      
      filpatcher.fillInterp(a_vel, oldCrseVel, newCrseVel,
                            crse_time_interp_coeff, vel_comp, dest_comp, 
                            num_comp);
    }
  }

  // need to set physical boundary conditions here
  basicVelBC* velBCPtr;
  if (s_nu > 0.0) {
    velBCPtr = m_physBCPtr->viscousVelBC();
  } else {
    velBCPtr = m_physBCPtr->tracingVelBC();
  }

  velBCPtr->applyBCs(a_vel, m_problem_domain, m_dx);

  delete velBCPtr;
  
  // should probably have exchange here
  a_vel.exchange(a_vel.interval());

}


// ------------------------------------------------------------------
void
AMRNavierStokes::velocity(LevelData<FArrayBox>& a_vel, Real a_time) const
{
  
  Interval velComps(0,SpaceDim-1);
  if (abs(a_time-m_time) < TIME_EPS) {
    m_vel_new_ptr->copyTo(velComps, a_vel, velComps);
  } else if (abs(a_time-(m_time-m_dt)) < TIME_EPS) {
    m_vel_old_ptr->copyTo(velComps, a_vel, velComps);
  } else {
    Real old_time = m_time - m_dt;
    timeInterp(a_vel, a_time, *m_vel_old_ptr, old_time, 
	       *m_vel_new_ptr, m_time, velComps);
  }

}


// ------------------------------------------------------------------
void
AMRNavierStokes::lambda(LevelData<FArrayBox>& a_lambda, Real a_time) const
{
  
  Interval lambdaComps(0,0);
  if (abs(a_time-m_time) < TIME_EPS) {
    m_lambda_new_ptr->copyTo(lambdaComps, a_lambda, lambdaComps);
  } else if (abs(a_time-(m_time-m_dt)) < TIME_EPS) {
    m_lambda_old_ptr->copyTo(lambdaComps, a_lambda, lambdaComps);
  } else {
    Real old_time = m_time - m_dt;
    timeInterp(a_lambda, a_time, *m_lambda_old_ptr, old_time, 
	       *m_lambda_new_ptr, m_time, lambdaComps);
  }

}


void
AMRNavierStokes::fillScalars(LevelData<FArrayBox>& a_scal,
			     Real a_time, const int a_comp) const
{
  // this will handle allocation as well...
  assert (!a_scal.isDefined());
  
  IntVect ghostVect(D_DECL(ADVECT_GROW, ADVECT_GROW, ADVECT_GROW));
  const DisjointBoxLayout& levelGrids = m_vel_new_ptr->getBoxes();

  a_scal.define(levelGrids, m_scal_new[a_comp]->nComp(), ghostVect);
  
  if (s_set_bogus_values) {
    DataIterator dit = a_scal.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      a_scal[dit()].setVal(s_bogus_value);
    }
  }  
  

  Interval scalComps = a_scal.interval();


  Real old_time = m_time - m_dt;
  // first copy on this level
  if (abs(a_time - old_time) < TIME_EPS) {
    m_scal_old[a_comp]->copyTo(scalComps, a_scal, scalComps);
  } else if (abs(a_time - m_time) < TIME_EPS) {
    m_scal_new[a_comp]->copyTo(scalComps, a_scal, scalComps);
  } else {
    // do linear interpolation in time
    timeInterp(a_scal, a_time, oldScal(a_comp), old_time,
	       newScal(a_comp), m_time, scalComps);
  }
  
  if (m_level > 0) {
    const DisjointBoxLayout& levelGrids = a_scal.getBoxes();
    
    // fill in coarse-fine BC data by conservative linear interp
    AMRNavierStokes& crseLevel = *crseNSPtr();
    
    LevelData<FArrayBox>& oldCrseScal = crseLevel.oldScal(a_comp);
    LevelData<FArrayBox>& newCrseScal = crseLevel.newScal(a_comp);
    const DisjointBoxLayout& crseGrids = oldCrseScal.getBoxes();
    const ProblemDomain& crseDomain = crseLevel.problemDomain();
    int nRefCrse = crseLevel.refRatio();
    
    Real crse_new_time = crseLevel.m_time;
    Real crse_dt = crseLevel.dt();
    Real crse_old_time = crse_new_time - crse_dt;
    Real crse_time_interp_coeff;
    
    // check for "essentially 0 or 1"
    if (abs(a_time - crse_old_time) < TIME_EPS) {
      crse_time_interp_coeff = 0.0;
    } else if (abs(a_time - crse_new_time) < TIME_EPS) {
      crse_time_interp_coeff = 1.0;
    } else {
      crse_time_interp_coeff = (a_time - crse_old_time)/crse_dt;
    }
    
    
    
    const IntVect& scalGrowVect = a_scal.ghostVect();
    int scalGrow = scalGrowVect[0];
    
    PiecewiseLinearFillPatch filpatcher(levelGrids, crseGrids, 
					a_scal.nComp(), crseDomain,
					nRefCrse, scalGrow);
    
    
    filpatcher.fillInterp(a_scal, oldCrseScal, newCrseScal,
			  crse_time_interp_coeff, 0,0,a_scal.nComp());
  }
  
  //  now do physical boundary conditions

  DomainGhostBC* thisBCPtr = m_physBCPtr->scalarTraceBC(a_comp);

  // loop over boxes
  const ProblemDomain& physDomain = problemDomain();
  DataIterator dit = a_scal.dataIterator();
  for (dit.reset(); dit.ok(); ++dit) {
    thisBCPtr->applyInhomogeneousBCs(a_scal[dit()], physDomain, m_dx);
  }
  
  delete thisBCPtr;

}
