/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// channelBC.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000

#include "channelBC.H"
#include "scalarDirichletBC.H"
#include "ExtrapBC.H"
#include "HOExtrapBC.H"
#include "PoissonBC.H"
#include "edgeGhostBC.H"
// this is included for scalar type enumeration
#include "AMRNavierStokes.H"
#include "CHANNELBCF_F.H"
#include "PROB_F_F.H"

#include "Timer.H"

extern Timer NST_Everything;
Timer TF_INITSCAL_channelBC("INITSCAL chBC",   NST_Everything, DIAG_TABLE4);

/*****************/
/*****************/

channelBC::channelBC() 
{

  // initialize to bogus values
  for (int dir=0; dir<SpaceDim; dir++) {
    m_loBC[dir] = bogusBC;
    m_hiBC[dir] = bogusBC;
  }

  m_maxInflowVel = 1.0e8;
  
  // now initialize BCs
  setBCs();

       

}

channelBC::~channelBC() 
{
}

PhysBCUtil*
channelBC::newPhysBCUtil() const
{
  channelBC* newBCPtr = new channelBC;
  newBCPtr->setMaxInflowVel(m_maxInflowVel);
  return static_cast<PhysBCUtil*>(newBCPtr);
}


DomainGhostBC*
channelBC::gradMacPressureBC() const
{
  DomainGhostBC* newBCPtr = new DomainGhostBC;
  
  // loop over directions
  for (int dir=0; dir<SpaceDim; ++dir) {
    SideIterator sit;
    for (sit.reset(); sit.ok(); ++sit) {
      int bctype = getBC(dir,sit());
      if (bctype == SolidWall) {
	HOExtrapBC thisBC(dir,sit());
	newBCPtr->setBoxGhostBC(thisBC);
      } else if (bctype == Inflow) {
	HOExtrapBC thisBC(dir,sit());
	newBCPtr->setBoxGhostBC(thisBC);
      } else if (bctype == Outflow) {
	ExtrapBC thisBC(dir,sit());
	newBCPtr->setBoxGhostBC(thisBC);
      } else if (bctype == noShear) {
	HOExtrapBC thisBC(dir,sit());
	//DircheletBC thisBC(dir,sit());
	newBCPtr->setBoxGhostBC(thisBC);
      } else {
	MayDay::Error("PhysBCUtil::basicPressureBC -- bad bctype");
      }
    } // end loop over sides
  } // end loop over directions
  
  return newBCPtr;
  
}

DomainGhostBC* 
channelBC::gradELambdaBC() const
{
  // think this should be the same as MAC bc (except homogeneous)
  DomainGhostBC* newBCPtr = gradMacPressureBC();
  return newBCPtr;
}

DomainGhostBC* 
channelBC::diffusiveBC(int a_scalarType) const
{
  bool isHomogeneous = false;
  
  DomainGhostBC* newBCPtr = new DomainGhostBC;

  for (int dir=0; dir<SpaceDim; dir++) {
    SideIterator sit;
    for (sit.reset(); sit.ok(); ++sit) {
      int bctype = getBC(dir,sit());
      if (bctype == Inflow) {
	scalarInflowChannelBC thisBC(dir,sit());
	thisBC.setScalarType(a_scalarType);
	thisBC.isHomogeneous(isHomogeneous);
	newBCPtr->setBoxGhostBC(thisBC);
      } else if (bctype == Outflow) {
	ExtrapBC thisBC(dir,sit());
	newBCPtr->setBoxGhostBC(thisBC);
      } else if (bctype == SolidWall) {
	// no diffusive flux across boundary
	NeumannBC thisBC(dir,sit());
	newBCPtr->setBoxGhostBC(thisBC);	 
      } else if (bctype == noShear) {
	NeumannBC thisBC(dir,sit());
	newBCPtr->setBoxGhostBC(thisBC);
      } else {
	MayDay::Error("channelBC::basicScalBC -- invalid bctype");
      }
    } // end loop over hi-lo
    
  } // end loop over directions
  
  return newBCPtr;
}
      


DomainGhostBC* 
channelBC::basicScalBC(bool a_isHomogeneous,
			  int a_scalType) const
{

  DomainGhostBC* newBCPtr = new DomainGhostBC;

  for (int dir=0; dir<SpaceDim; dir++) {
    SideIterator sit;
    for (sit.reset(); sit.ok(); ++sit) {
      int bctype = getBC(dir,sit());
      if (bctype == Inflow) {
	scalarInflowChannelBC thisBC(dir,sit());
	thisBC.setScalarType(a_scalType);
	thisBC.isHomogeneous(a_isHomogeneous);
	newBCPtr->setBoxGhostBC(thisBC);
      } else if (bctype == Outflow) {
	ExtrapBC thisBC(dir,sit());
	newBCPtr->setBoxGhostBC(thisBC);
      } else if (bctype == SolidWall) {
	// all other scalars are extrap BC
	ExtrapBC thisBC(dir,sit());
	newBCPtr->setBoxGhostBC(thisBC);	
      } else if (bctype == noShear) {
	ExtrapBC thisBC(dir,sit());
	newBCPtr->setBoxGhostBC(thisBC);
      } else {
	MayDay::Error("channelBC::basicScalBC -- invalid bctype");
      }
    } // end loop over hi-lo
    
  } // end loop over directions
  
  return newBCPtr;
}


void
channelBC::basicCCVelBC(DomainGhostBC& a_newBC,
			bool a_isHomogeneous,
			bool a_isViscous,
			int a_comp, Interval& a_interval) const
{
  // loop over directions
  for (int dir=0; dir<SpaceDim; dir++) {
    // loop over hi-lo
    SideIterator sit;
    for (sit.begin(); sit.ok(); ++sit) {
      int bctype = getBC(dir,sit());
      
      if (bctype == SolidWall) {
      // normal velocity BC's are always no-flow
	if (dir == a_comp) {
	  DircheletBC thisBC(dir,sit(), a_interval);
	  a_newBC.setBoxGhostBC(thisBC);
	} else {
	  // tangential BCs are no-slip if viscous, extrap if invisice
	  if (a_isViscous) {
	    DircheletBC thisBC(dir,sit(), a_interval);
	    a_newBC.setBoxGhostBC(thisBC);
	  } else {
	    ExtrapBC thisBC(dir,sit(),a_interval);
	    a_newBC.setBoxGhostBC(thisBC);
	  }

	} // end if tangential
      } else if (bctype == Inflow) {
	// normal velocity at inflow boundary
	if (dir == a_comp) {
	  NormalInflowVelBC thisBC(dir,sit(),a_interval);
	  thisBC.isHomogeneous(a_isHomogeneous);
	  thisBC.setBCVal(1.0);
	  a_newBC.setBoxGhostBC(thisBC);
	} else {
	  // tangential vel at inflow is 0 for this problem
	  DircheletBC thisBC(dir,sit(),a_interval);
	  a_newBC.setBoxGhostBC(thisBC);
	}
      } else if (bctype == Outflow) {
	NeumannBC thisBC(dir,sit(), a_interval);
	a_newBC.setBoxGhostBC(thisBC);
      } else if (bctype == noShear) {
	if (dir == a_comp) {
	  DircheletBC thisBC(dir,sit(),a_interval);
	  a_newBC.setBoxGhostBC(thisBC);
	} else {
	  // tangential BC are neumann
	  NeumannBC thisBC(dir,sit(),a_interval);
	  a_newBC.setBoxGhostBC(thisBC);
	}
      } else {
	MayDay::Error("PhysBCUtil::basicCCVelBC -- bad bctype");
      }

    } // end iteration over sides
  } // end iteration over directions
}
	



void
channelBC::setMaxInflowVel(Real a_maxInflowVel)
{
  m_maxInflowVel = a_maxInflowVel;
}

Real
channelBC::maxInflowVel() const
{
  return m_maxInflowVel;
}


void
channelBC::setBCs() 
{
  

  ParmParse ppBC("physBC");

  vector<int> tempBC(SpaceDim);

  ppBC.getarr("lo", tempBC, 0, SpaceDim);
  for (int dir=0; dir<SpaceDim; dir++) {
    m_loBC[dir] = tempBC[dir];
  }

  ppBC.getarr("hi", tempBC, 0, SpaceDim);

  for (int dir=0; dir<SpaceDim; dir++) {
    m_hiBC[dir] = tempBC[dir];
  }

  ppBC.query("maxInflowVel", m_maxInflowVel);

}


void
channelBC::computeBoundaryDt(Real& a_dt, Real a_cfl, Real a_Dx) const
{

  Real maxInflowDt = a_cfl*a_Dx/m_maxInflowVel;
  
  if (maxInflowDt < a_dt) a_dt = maxInflowDt;

}


/// functions for normal component of inflowVel

NormalInflowVelBC::NormalInflowVelBC() : InflowVelBC()
{
}


NormalInflowVelBC::~NormalInflowVelBC() 
{
}

NormalInflowVelBC::NormalInflowVelBC(int dir, Side::LoHiSide side) 
  : InflowVelBC(dir,side)
{
}

NormalInflowVelBC::NormalInflowVelBC(int dir, Side::LoHiSide side,
				     const Interval& a_comps) 
  : InflowVelBC(dir,side,a_comps)
{
}

BoxGhostBC*
NormalInflowVelBC::new_boxghostbc() const
{
  NormalInflowVelBC* newop = new NormalInflowVelBC();
  if (newop == NULL) {
    MayDay::Error("Out of memory in NormalInflowVelBC::new_boxghostbc");
  }
  newop->isHomogeneous(isHomogeneous());
  newop->setBCVal(BCVal());
  return static_cast<BoxGhostBC*> (newop);
}

void
NormalInflowVelBC::fillBCValues(FArrayBox& a_neumfac,
				FArrayBox& a_dircfac,
				FArrayBox& a_inhmval,
				Real a_dx,
				const Box& a_domain) const
{

  assert(!a_neumfac.box().isEmpty());
  assert(!a_dircfac.box().isEmpty());
  assert(!a_inhmval.box().isEmpty());
  a_neumfac.setVal(0.0);
  a_dircfac.setVal(1.0);

  // for now, set inflow velocity to be one.
  if (isHomogeneous()) {
    a_inhmval.setVal(0.0);
  } else {
    for (int comp=m_components.begin(); comp <= m_components.end(); comp++){
      FORT_CHANNELNORMINFLOWBC(CHF_FRA1(a_inhmval, comp),
			     CHF_REAL(a_dx),
			     CHF_CONST_INT(m_direction),
			     CHF_BOX(a_domain));    
    }
  }  
}


scalarInflowChannelBC::scalarInflowChannelBC()
  : scalarInflowBC()
{
}

scalarInflowChannelBC::scalarInflowChannelBC(int dir, Side::LoHiSide sd)
  : scalarInflowBC(dir,sd)
{
}

scalarInflowChannelBC::scalarInflowChannelBC(int dir, Side::LoHiSide sd,
					     const Interval& a_comps)
  : scalarInflowBC(dir,sd,a_comps)
{
}

///
scalarInflowChannelBC::~scalarInflowChannelBC()
{
}


BoxGhostBC*
scalarInflowChannelBC::new_boxghostbc() const
{
  scalarInflowChannelBC* newop = new scalarInflowChannelBC();
  if(newop == NULL)
    {
      MayDay::Error("Out of Memory in scalarInflowChannelBC::new_boxghostbc");
    }
  
  newop->setScalarType(scalarType());
  newop->isHomogeneous(isHomogeneous());
  return static_cast<BoxGhostBC*>(newop);
}
  

void
scalarInflowChannelBC::fillBCValues(FArrayBox& a_neumfac,
				    FArrayBox& a_dircfac,
				    FArrayBox& a_inhmval,
				    Real a_dx,
				    const Box& a_domain) const
{
  // this will need to be reworked a little.
  assert(!a_neumfac.box().isEmpty());
  assert(!a_dircfac.box().isEmpty());
  assert(!a_inhmval.box().isEmpty());
  a_neumfac.setVal(0.0);
  a_dircfac.setVal(1.0);

  if (isHomogeneous()) {
    a_inhmval.setVal(0.0);
  } else {
    // can reuse the initscal function here
    int scalComp = scalarType();

    int isEdgeCentered=0;
    int hiLo = 0;
    Box testBox = a_inhmval.box();
    for (int dir=0; dir<SpaceDim; dir++) {
      Box testBoxLo = testBox;
      testBoxLo.shift(dir,1);
      if (testBoxLo.intersects(a_domain)) {
	isEdgeCentered = dir+1;
	hiLo =-1;
	a_inhmval.shiftHalf(dir,1);				    
      }
      Box testBoxUpper = testBox;
      testBoxUpper.shift(dir,-1);
      if (testBoxUpper.intersects(a_domain)) {
	isEdgeCentered = dir+1;
	hiLo = 1;
	a_inhmval.shiftHalf(dir,-1);
      }
    }


    // if this is multicomponent, then 
    //timercomment//TF_INITSCAL_channelBC.start();
    FORT_INITSCAL(CHF_FRA1(a_inhmval,0),
		  CHF_REAL(a_dx), CHF_INT(scalComp),
		  CHF_BOX(a_domain),
		  CHF_INT(isEdgeCentered)); 
    //timercomment//TF_INITSCAL_channelBC.stop();
   
    if (isEdgeCentered != 0) {
      int dir = isEdgeCentered -1;
      a_inhmval.shiftHalf(dir,hiLo);
    }
  }
  
}


basicEdgeVelBC*
channelBC::basicECVelBC(bool a_isHomogeneous, bool a_isViscous,
			bool a_isSingleComponent) const
{
  
  basicEdgeVelBC* newBCPtr = new basicEdgeVelBC;
  // at the moment, only single component defined
  assert (a_isSingleComponent == true);
  newBCPtr->isSingleComponent(a_isSingleComponent);

  // loop over component directions
  for (int comp=0; comp<SpaceDim; comp++) {
    DomainGhostBC thisCompBC;
    Interval compInterval(0,0);

    IndexType edgeVelType;
    // first set all directions to be cell-centered
    edgeVelType.clear();
    // then set normal direction to be edge-centered
    edgeVelType.set(comp);
    
    // loop over directions
    for (int dir=0; dir<SpaceDim; dir++) {
      SideIterator sit;
      for (sit.begin(); sit.ok(); ++sit) {
	int bctype = getBC(dir,sit());
	
	// solid-wall BCs
	if (bctype == SolidWall) {
	  // normal BC's are always no-flow
	  if (dir == comp) {
	    edgeDirichletBC thisBC(dir,sit(), compInterval);
	    thisBC.ixType(edgeVelType);
	    thisBC.setBCVal(0.0);
	    thisCompBC.setBoxGhostBC(thisBC);
	  } else {
	    // tangential BC's are no-slip if viscous, extrap if inviscid
	    if (a_isViscous) {
	      DircheletBC thisBC(dir,sit(),compInterval);
	      thisCompBC.setBoxGhostBC(thisBC);
	    } else {
	      ExtrapBC thisBC(dir,sit(),compInterval);
	      thisCompBC.setBoxGhostBC(thisBC);
	    }
	  } // end if tangential comp
	} else if (bctype == Inflow) {
	  //  this will have to be changed for a more complicated 
	  // inflow profile, but it will work for now
	  edgeDirichletBC thisBC(dir,sit(),compInterval);
	  thisBC.ixType(edgeVelType);
	  if (dir == comp) {
	    // normal velocity is uniform
	    if (a_isHomogeneous) {
	      thisBC.setBCVal(0.0);
	    } else {	      
	      thisBC.setBCVal(1.0);
	    }
	  } else {
	    // no tangential velocity at inflow
	    thisBC.setBCVal(0.0);
	  }
	  thisCompBC.setBoxGhostBC(thisBC);
	} else if (bctype == Outflow) {
	  // this is set to whatever it is set to by MAC
	  NoOpBC thisBC(dir,sit(),compInterval);
	  thisCompBC.setBoxGhostBC(thisBC);
	} else if (bctype == noShear) {
	  if (dir == comp) {
	    // normal vel is no-flow
	    edgeDirichletBC thisBC(dir,sit(),compInterval);
	    thisBC.ixType(edgeVelType);
	    thisBC.setBCVal(0.0);
	    thisCompBC.setBoxGhostBC(thisBC);
	  } else {
	    // do nothing in tangential case
	    NoOpBC thisBC(dir,sit(),compInterval);
	    thisCompBC.setBoxGhostBC(thisBC);
	  }
	} else {
	  MayDay::Error("PhysBCUtil::basicECVelBC -- bad bctype");
	}

      } // end loop over hi-lo
    } // end loop over directions

    newBCPtr->setBCType(thisCompBC,comp);
    
  } // end loop over velocity components

  return newBCPtr;
}
