/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// Mask.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


#include "Mask.H"
#include "LayoutIterator.H"

// -------------------------------------------------------------
void
Mask::buildMask(BaseFab<int>& a_mask, const Box& a_dProblem, 
		const BoxLayout& a_grids, const BoxLayout* a_fineGridsPtr,
		int a_nRefFine) 
{
  ProblemDomain physdomain(a_dProblem);
  buildMask(a_mask, physdomain, a_grids, a_fineGridsPtr, 
            a_nRefFine);
}

// -------------------------------------------------------------
void
Mask::buildMask(BaseFab<int>& a_mask, const ProblemDomain& a_dProblem, 
		const BoxLayout& a_grids, const BoxLayout* a_fineGridsPtr,
		int a_nRefFine) 
{
  // first set entire box to Physical BC
  a_mask.setVal(maskPhysical);

  // now set all of domain interior to coarse
  Box domainInterior(a_mask.box());
  domainInterior &= a_dProblem;
  a_mask.setVal(maskCoarse,domainInterior,0);

  // now loop over this level's boxes and set them to "copy"
  LayoutIterator lit = a_grids.layoutIterator();
  
  for (lit.reset(); lit.ok(); ++lit) {
    Box intersectBox = a_grids.get(lit());
    intersectBox &= a_mask.box();
    if (!intersectBox.isEmpty()) {
      a_mask.setVal(maskCopy,intersectBox,0);
    }
  }

  // if finer grids exist, set them to "covered"
  if (a_fineGridsPtr != NULL) {
    assert (a_nRefFine > 1);
    
    LayoutIterator litFine = a_fineGridsPtr->layoutIterator();
    for (litFine.reset(); litFine.ok(); ++litFine) {
      Box coarsenedBox(a_fineGridsPtr->get(litFine()));
      coarsenedBox.coarsen(a_nRefFine);
      coarsenedBox &= a_mask.box();
      if (!coarsenedBox.isEmpty()) {
	a_mask.setVal(maskCovered,coarsenedBox,0);
      }
    }
  }

}



// -------------------------------------------------------------
void 
Mask::buildMasks(LevelData<BaseFab<int> >& a_masks, const Box& a_dProblem,
		 const BoxLayout& a_grids, const BoxLayout* a_fineGridsPtr,
		 int a_nRefFine)
{
  ProblemDomain physdomain(a_dProblem);
  buildMasks(a_masks, a_dProblem, a_grids, a_fineGridsPtr, a_nRefFine);
}


// -------------------------------------------------------------
void 
Mask::buildMasks(LevelData<BaseFab<int> >& a_masks, 
                 const ProblemDomain& a_dProblem,
		 const BoxLayout& a_grids, const BoxLayout* a_fineGridsPtr,
		 int a_nRefFine)
{

  // simply iterate over boxes and call single-basefab function
  DataIterator dit = a_masks.dataIterator();
  
  for (dit.reset(); dit.ok(); ++dit) {
    buildMask(a_masks[dit()], a_dProblem, a_grids, a_fineGridsPtr, 
	      a_nRefFine);
  }

}
