/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// computeNorm.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


#include "computeNorm.H"
#include "LayoutIterator.H"
#include "BoxLayoutData.H"
#ifdef MPI
#include <mpi.h>
#endif

#include "Timer.H"
//timercomment//extern Timer  NST_Everything;
//timercomment//Timer NST_MPI_normReduce("ns norm Reduce", NST_Everything);

// --------------------------------------------------------------
Real
computeNorm(const Vector<LevelData<FArrayBox>* >& a_phi, 
	    const Vector<int>& a_nRefFine,
	    const Real a_dxCrse, const Interval a_comps, 
	    const int p, const int lBase)
{

  Real dxLevel = a_dxCrse;
  int numLevels = a_phi.size();
  Real norm, normLevel;

  norm = 0.0;

  // loop over levels  
  for (int lev=lBase;  lev< numLevels; lev++) {
    //in case there are extra levels which are not defined
    if (a_phi[lev] != NULL) {
      assert(a_phi[lev]->isDefined());
      
      LevelData<FArrayBox>& thisPhi = *(a_phi[lev]);
      const DisjointBoxLayout* finerGridsPtr = NULL;
      
      if (lev < numLevels-1) {
	finerGridsPtr = &(a_phi[lev+1]->getBoxes());
      }

      

      normLevel = computeNorm(thisPhi, finerGridsPtr, a_nRefFine[lev], 
			    dxLevel, a_comps, p);
      if (p == 2) normLevel = normLevel*normLevel;
      if (p != 0) {

	norm += normLevel;
      } else if (normLevel > norm) { 	
	norm = normLevel;
      }
      
    }

    // update  dxLevel
    dxLevel = dxLevel/a_nRefFine[lev];
    
  }
  
  // shouldn't need to do broadcast/gather thing 

  // now do sqrt, etc
  if (p==2) norm = sqrt(norm);
  return norm;
}

// ------------------------------------------------------------
Real 
computeNorm(const LevelData<FArrayBox>& a_phi, 
	    const DisjointBoxLayout* a_finerGridsPtr,
	    const int a_nRefFine, const Real a_dx, 
	    const Interval a_comps, const int p)
{

  Real normLevel;
  normLevel = 0.0;

  const DisjointBoxLayout& levelGrids = a_phi.getBoxes();

  LevelData<FArrayBox> temp(levelGrids, a_comps.size());

  Interval tempComps(0,a_comps.size()-1);
  
  DataIterator dit = a_phi.dataIterator();

  for (dit.reset(); dit.ok(); ++dit) {
    temp[dit()].copy(temp[dit()].box(), tempComps,
                     temp[dit()].box(), a_phi[dit()], a_comps);

    if (a_finerGridsPtr != NULL) {  
      LayoutIterator litFine = a_finerGridsPtr->layoutIterator();
      
      // now loop over fine boxes and set covered regions to 0
      for (litFine.reset(); litFine.ok(); ++litFine) {
	Box coveredBox(a_finerGridsPtr->get(litFine()));
	coveredBox.coarsen(a_nRefFine);
	coveredBox &= temp[dit()].box();
	
	if (!coveredBox.isEmpty()) {
	  temp[dit()].setVal(0.0, coveredBox, 0, tempComps.size());
	}
      }  // end loop over fine-grid boxes

    } // end if there is a finer level
    
  } // end loop over this level's grids
  
  // compute unscaled norm
  normLevel = norm(temp, tempComps, p);

  if (p!=0) {
    Real scale = pow(a_dx,(SpaceDim/p));
    normLevel *= scale;
  }
  
  return normLevel;

}


      
 /// extension of the norm(...) function for FluxBoxes
Real norm(const BoxLayoutData<FluxBox>& a_layout,
	  const Interval& interval, const int& p)
{
  const BoxLayout& validRegion = a_layout.boxLayout();
  DataIterator dit = validRegion.dataIterator();
  Real norm = 0;
  if (p == 0) {
    for (dit.reset(); dit.ok(); ++dit) {
      const FluxBox& thisFlux = a_layout[dit()];	
      for (int dir=0; dir<SpaceDim; dir++) {
	norm = Max(norm, thisFlux[dir].norm(validRegion[dit()],
					    0, interval.begin(),
					    interval.size()));
      }
    }
#ifdef MPI
    Real recv;
    //timercomment//NST_MPI_normReduce.start();
    int result = MPI_Allreduce(&norm, &recv, 1, MPI_CH_REAL,
			       MPI_MAX, Chombo_MPI::comm);
    //timercomment//NST_MPI_normReduce.stop();
      if(result != MPI_SUCCESS){ //bark!!!
	MayDay::Error("sorry, but I had a communcation error on norm");
      }
      norm = recv;
#     endif
    }
  else if(p == 1) // abs sum norm
    {
      for(dit.reset();dit.ok(); ++dit)
	{
	  const FluxBox& thisFlux = a_layout[dit()];
	  for (int dir=0; dir<SpaceDim; ++dir) {
	    norm += thisFlux[dir].norm(validRegion[dit()],
				       1,
				       interval.begin(),
				       interval.size());
	  }
	}
#ifdef MPI
      Real recv;
      //timercomment//NST_MPI_normReduce.start();
      int result = MPI_Allreduce(&norm, &recv, 1, MPI_CH_REAL,
				 MPI_SUM, Chombo_MPI::comm);
      //timercomment//NST_MPI_normReduce.stop();
      if(result != MPI_SUCCESS){ //bark!!!
	MayDay::Error("sorry, but I had a communcation error on norm");
      }
      norm = recv;
#endif
    } else {
    for(dit.reset();dit.ok(); ++dit)
      {
	const FluxBox& thisFlux = a_layout[dit()];
	for (int dir=0; dir<SpaceDim; dir++) {
	  norm+=thisFlux[dir].sumPow(validRegion[dit()],
				     p,
				     interval.begin(),
				     interval.size());
	}
      }
#ifdef MPI
    Real recv;
    //timercomment//NST_MPI_normReduce.start();
    int result = MPI_Allreduce(&norm, &recv, 1, MPI_CH_REAL,
			       MPI_SUM, Chombo_MPI::comm);
    //timercomment//NST_MPI_normReduce.stop();
    if(result != MPI_SUCCESS){ //bark!!!
      MayDay::Error("sorry, but I had a communcation error on norm");
    }
    norm = recv;
#     endif
    Real invpwr = 1.0/p;
    norm = pow(norm, invpwr);
    }

  return norm;
}

      
   
